% Written by Remo Kretschmann

function x = draw_mixed(a,b,c)
% DRAW_MIXED draws a sample of a real random variable x with probability density p(x) = exp(-a*x^2 - abs(b*x - c))/Z with a, b > 0 using the inverse cumulative distribution method.
%
% For details on the implementation, see section 6.4 of
% Remo Kretschmann: Nonparametric Bayesian Inverse Problems with Laplacian Noise (2019)
% https://doi.org/10.17185/duepublico/70452

sqrt_a = sqrt(a);
gamma = b/(2*sqrt_a);
x_0 = c/b; % Kink of the pdf

% The points where erfcx is evaluated.
gamma_minus = gamma - sqrt_a*c/b;
gamma_plus = gamma + sqrt_a*c/b;

log_2 = log(2); % Natural logarithm
log_pi = log(pi);

% Case both gamma_plus >= 0 and gamma_minus >= 0
if gamma_plus >= 0 && gamma_minus >= 0
	omega = erfcx(gamma_plus) + erfcx(gamma_minus);
	r = rand(); % Draw from uniform distribution on [0,1]

	% left side of the distribution
	w = log(r) - gamma_minus^2 + log(omega);
	if w < -680 && abs(log(gamma_minus)) > 3
	    theta = - log_pi - log(-w);
	    v = (-theta - 2);
	    s = 2/(theta - 2 * w);
	    a2 = 0.125 * v;
	    a3 = - 0.03125 * (v^2 + 6*v - 6);
	    a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
	    erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
	elseif abs(log(gamma_minus)) > 3
	    erfcinv_z = erfcinv(exp(w));
	else
	    erfcinv_z = erfcinv(r * exp(-gamma_minus^2) * omega);
	end
	x = (gamma - erfcinv_z)/sqrt_a;

	% right side of the distribution
	if x > x_0
	    w = log(1-r) - gamma_plus^2 + log(omega);
	    if w < -680 && abs(log(gamma_plus)) > 3
		theta = - log_pi - log(-w);
		v = (-theta - 2);
		s = 2/(theta - 2 * w);
		a2 = 0.125 * v;
		a3 = - 0.03125 * (v^2 + 6*v - 6);
		a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
		erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
	    elseif abs(log(gamma_plus)) > 3
		erfcinv_z = erfcinv(exp(w));
	    else
		erfcinv_z = erfcinv((1-r) * exp(-gamma_plus^2) * omega);
	    end
	    x = (erfcinv_z - gamma)/sqrt_a;
	end

% Case gamma_plus >= 0 and gamma_minus < 0
elseif gamma_minus < 0
	omega = erfcx(gamma_plus) - erfcx(-gamma_minus);
        r = rand();

        % left side of the distribution
        w = log(r) + log_2 ...
            + log(1 + sign(omega)*exp(log(abs(omega)) - gamma_minus^2 - log_2));
        if w < -680 && max(abs(log([gamma_plus,-gamma_minus]))) > 3
            theta = - log_pi - log(-w);
            v = (-theta - 2);
            s = 2/(theta - 2 * w);
            a2 = 0.125 * v;
            a3 = - 0.03125 * (v^2 + 6*v - 6);
            a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
            erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
        elseif max(abs(log([gamma_plus,-gamma_minus]))) > 3
            erfcinv_z = erfcinv(exp(w));
        else
            erfcinv_z = erfcinv(r * (exp(-gamma_minus^2) * omega + 2));
        end
        x = (gamma - erfcinv_z)/sqrt_a;

        % right side of the distribution
        if x > x_0
            w = log(1-r) + log_2 - 2*c ...
                + log(1 + sign(omega)*exp(log(abs(omega)) - gamma_plus^2 - log_2 + 2*c));
            if w < -680 && abs(log(gamma_plus)) > 3
                theta = - log_pi - log(-w);
                v = (-theta - 2);
                s = 2/(theta - 2 * w);
                a2 = 0.125 * v;
                a3 = - 0.03125 * (v^2 + 6*v - 6);
                a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
                erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
            elseif abs(log(gamma_plus)) > 3
                erfcinv_z = erfcinv(exp(w));
            else
                erfcinv_z = erfcinv((1-r) * (exp(-gamma_plus^2) * omega + 2*exp(-2*c)));
            end
            x = (erfcinv_z - gamma)/sqrt_a;
        end

% Case gamma_plus < 0 and gamma_minus >= 0
else
	omega = erfcx(-gamma_plus) - erfcx(gamma_minus);
        r = rand();

        % left side of the distribution
        w = log(r) + log_2 + 2*c ...
            + log(1 - sign(omega)*exp(log(abs(omega)) - gamma_minus^2 - log_2 - 2*c));
        if w < -680 && abs(log(gamma_minus)) > 3
            theta = - log_pi - log(-w);
            v = (-theta - 2);
            s = 2/(theta - 2 * w);
            a2 = 0.125 * v;
            a3 = - 0.03125 * (v^2 + 6*v - 6);
            a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
            erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
        elseif abs(log(gamma_minus)) > 3
            erfcinv_z = erfcinv(exp(w));
        else
            erfcinv_z = erfcinv(r * (-exp(-gamma_minus^2) * omega + 2*exp(2*c)));
        end
        x = (gamma - erfcinv_z)/sqrt_a;

        % right side of the distribution
        if x > x_0
            w = log(1-r) + log_2 ...
                + log(1 - sign(omega)*exp(log(abs(omega)) - gamma_plus^2 - log_2));
            if w < -680 && max(abs(log([-gamma_plus,gamma_minus]))) > 3
                theta = - log_pi - log(-w);
                v = (-theta - 2);
                s = 2/(theta - 2 * w);
                a2 = 0.125 * v;
                a3 = - 0.03125 * (v^2 + 6*v - 6);
                a4 = 1/384 * (4*v^3 + 27*v^2 + 108*v - 300);
                erfcinv_z = 1/sqrt(s) + a2 * s^(3/2) + a3 * s^(5/2) + a4 * s^(7/2);
            elseif max(abs(log([-gamma_plus,gamma_minus]))) > 3
                erfcinv_z = erfcinv(exp(w));
            else
                erfcinv_z = erfcinv((1-r) * (-exp(-gamma_plus^2) * omega + 2));
            end
            x = (erfcinv_z - gamma)/sqrt_a;
        end

end

end
